////////////////////////////
// JBEAM File Transformer //
////////////////////////////

///////////////////
// Include Files //
///////////////////
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	"ErrorCodes.h"
#include	"FileNames.h"
#include	"Coord_Trans.h"
#include	"JBEAM_Transformer.h"

////////////////////////
// Parser State Flags //
////////////////////////
#define			flagQuote	(0x01)	//In quoted section
#define			flagNodes	(0x02)	//In "nodes" section
#define			flagCmnt1	(0x04)	//In single-line comment
#define			flagCmnt2	(0x08)	//In multi-line comment
#define			flagNot1st	(0x10)	//Subsequent (not first) parameter set
#define			flagDEBUG	(0x80)	//DEBUG TEST FLAG

//////////////////////////////
// Double-bracketed section //
//    linked-list stuff     //
//////////////////////////////
typedef enum
{
	b2Type_FieldData = 0,
	b2Type_Comment,
	b2Type_Delimiter
}
B2LinkTypeEnum;
typedef struct B2LinkType
{
	B2LinkType*		pPrev;
	char*			buf;
	B2LinkTypeEnum	type;
	B2LinkType*		pNext;
}
B2LinkType,
*B2LinkPtr,
**B2LinkHndl;

//////////////////
// Parser State //
//////////////////
typedef struct
{
	//Input parameters
	const char*		sRotord;
	double			xRot;
	double			yRot;
	double			zRot;
	double			xMov;
	double			yMov;
	double			zMov;

	//Bracket & brace state "[{" etc; make 128 B
	char*			sState;

	//Quoted parameters, comments, and "nodes" section
	unsigned int	iState;

	//Buffer; Make this 4 KB
	char*			buf;

	//Output file, for writing
	FILE*			fDst;

	//Parameter section links
	B2LinkPtr		pLinks;
	B2LinkHndl		hNextLink;

	//Parameter positions within parameter set
	char			cIdx[5];

	//Last character received
	char			lastChar;
}
ParserStateType,
*ParserStatePtr;

/////////////////
// Append link //
/////////////////
B2LinkPtr				AppendLink(
const ParserStatePtr	s)
{
	//Allocate new link
	*s->hNextLink = B2LinkPtr(malloc(sizeof(B2LinkType)));
	B2LinkPtr link = *s->hNextLink;
	if (link == NULL)
	{
		printf("Cannot append link\r\n");
		return(NULL);
	}
	link->buf = NULL;

	//Set next link loc
	s->hNextLink = &link->pNext;

	//This is the only link
	if (link == s->pLinks)
	{
		link->pNext = link;
		link->pPrev = link;
	}

	//Already have other links
	else
	{
		//Connect previous link to new link
		//NOTE: Previous link's "pNext" already pointing to new link
		link->pPrev = s->pLinks->pPrev;

		//Connect next link to new link
		link->pNext = s->pLinks;
		s->pLinks->pPrev = link;
	}

	//Return new link, if any
	return(link);
}

/////////////////
// Remove link //
/////////////////
void					RemoveLink(
const ParserStatePtr	s,
const B2LinkPtr			link)
{
	//Validate link
	if (link == NULL) return;

	//This is only link in chain
	int	onlyLink = (link->pPrev == link) ? 1 : 0;

	//Next link loc pointing at this link
	if (s->hNextLink == &link->pNext)
	{
		//This is only link in chain
		if (onlyLink)
			s->hNextLink = &s->pLinks;

		//Got other links in chain
		else
			s->hNextLink = &link->pPrev->pNext;
	}

	//Only link; No other links to connect
	if (onlyLink)
		s->pLinks = NULL;

	//Got other links
	else
	{
		//Connect previous link to next link
		link->pPrev->pNext = link->pNext;

		//Connect next link to previous link
		link->pNext->pPrev = link->pPrev;

		//Move starting link
		if (s->pLinks == link)
			s->pLinks = link->pNext;
	}

	//Free link mem
	free(link);
}

///////////////////////
// Clear linked list //
///////////////////////
void					ClearLinkedList(
const ParserStatePtr	s)
{
	while (s->pLinks != NULL) RemoveLink(s, s->pLinks);
}

////////////////////////////
// Get link buffer bounds //
////////////////////////////
int						GetLinkBufBounds(
const ParserStatePtr	s,
const B2LinkPtr			link,
char**					hStart,
char**					hStop)
{
	//Validate link pointer
	if (link == NULL) return(0);

	//Get start and end of delimiter
	*hStart = link->buf;
	*hStop = (s->hNextLink == &link->pNext) ? (*hStart) + strlen(*hStart) : link->pNext->buf;

	//Bail
	return(1);
}

/////////////////////////
// Delimiter has comma //
/////////////////////////
int						DelHasComma(
const ParserStatePtr	s,
const B2LinkPtr			link)
{
	//Validate link pointer and type
	if (link == NULL) return(0);
	if (link->type != b2Type_Delimiter) return(0);

	//Get start and end of delimiter
	char* pStart = NULL;
	char* pStop = NULL;
	if (GetLinkBufBounds(s, link, &pStart, (char** const) &pStop) == 0) return(0);

	//Search for comma
	int		gotComma = 0;
	char*	pHere = pStart;
	while (pHere < pStop && gotComma == 0)
	{
		if (*pHere == ',') gotComma = 1;
		pHere++;
	}
	return(gotComma);
}

///////////////////////////
// Clock more characters //
///////////////////////////
int						ClockChars(
const ParserStatePtr	s,
char					c,
B2LinkTypeEnum			lt)
{
	//Vars
	int			rtn = gecNoError;
	B2LinkPtr	link = NULL;

	//Current offset in buffer
	long		off = long(strlen(s->buf) - 1);

	//No links yet
	if (s->pLinks == NULL)
	{
		//Need new link
		link = AppendLink(s);
		if (link == NULL)
			rtn = gecMemAllocErr;
		else
		{
			link->type = lt;
			link->buf = s->buf + off;
		}
	}

	//Got links
	else
	{
		//Get current link
		link = s->pLinks->pPrev;

		//Different link type, or delimiter already with comma and received comma
		if (link->type != lt || (link->type == b2Type_Delimiter && DelHasComma(s, link)))
		{
			//Need new link
			link = AppendLink(s);
			if (link == NULL)
				rtn = gecMemAllocErr;
			else
			{
				link->type = lt;
				link->buf = s->buf + off;
			}
		}
	}

	//Bail
	return(rtn);
}

//////////////////////////////////////
// Get last bounding char collected //
//////////////////////////////////////
char					GetLastBoundChar(
const ParserStatePtr	s)
{
	//Return char
	char	cLast = '\0';

	//Got bounding chars?
	int		sLen = (int)strlen(s->sState);
	if (sLen > 0)
	{
		//Get last char
		cLast = s->sState[sLen - 1];
	}

	//Bail
	return(cLast);
}

///////////////////////////////
// Delete last bounding char //
///////////////////////////////
void					DelLastBoundChar(
const ParserStatePtr	s)
{
	int		sLen = (int)strlen(s->sState);
	if (sLen > 0) s->sState[sLen - 1] = '\0';
}

///////////////////////////////////////////////////////
// Bounding chars are right for paramater collection //
///////////////////////////////////////////////////////
int						BoundsRightForCol(
const ParserStatePtr	s)
{
	//Return var
	int		yep = 0;

	if (charCount(s->sState, '[') >= 2 &&
		charCount(s->sState, '{') >= 2)
		yep = 1;

	//Bail
	return(yep);
}

///////////////////////////////////
// In parameter collection state //
///////////////////////////////////
int						InParmCol(
const ParserStatePtr	s)
{
	//Return var
	int		yep = 0;

	if ((s->iState & flagNodes) &&
		BoundsRightForCol(s) != 0)
		yep = 1;

	//Bail
	return(yep);
}

///////////////
// Save char //
///////////////
int						SaveChar(
const ParserStatePtr	s,
char					c,
B2LinkTypeEnum			lt)
{
	//Return var
	int	rtn = gecNoError;

	//2nd+ level bracket in "nodes" section; Stash for later parameter processing
	if (InParmCol(s))
	{
		strcatc(s->buf, c);
		ClockChars(s, c, lt);
	}

	//Not processing parameters; Dump directly to file
	else
	{
		int	fStat = fputc(c, s->fDst);
		if (fStat == EOF)
		{
			printf("Error writing to destination file\r\n");
			rtn = gecFileWrite;
		}
	}

	//Bail
	return(rtn);
}

/////////////////////////////////
// Process first parameter set //
/////////////////////////////////
int						ProcParmHdr(
const ParserStatePtr	s)
{
	//Return var
	int		rtn = gecNoError;

	//Loop through links
	char*	pStart = NULL;
	char*	pStop = NULL;
	B2LinkHndl	link = &s->pLinks;
	while (link != s->hNextLink && rtn == gecNoError)
	{
		if ((*link)->type == b2Type_FieldData)
		{
			//Find field tags and determine their order
			if (GetLinkBufBounds(s, (*link), &pStart, &pStop) != 0)
			{
				if (cistrncmp((*link)->buf, "\"id\"", 4) != NULL)
				{
					if (strchr(s->cIdx, 'i') == NULL)
						strcat(s->cIdx, "i");
					else
						rtn = gecJbeamFormatErr;
				}
				else if (cistrncmp((*link)->buf, "\"posX\"", 6) != NULL)
				{
					if (strchr(s->cIdx, 'x') == NULL)
						strcat(s->cIdx, "x");
					else
						rtn = gecJbeamFormatErr;
				}
				else if (cistrncmp((*link)->buf, "\"posY\"", 6) != NULL)
				{
					if (strchr(s->cIdx, 'y') == NULL)
						strcat(s->cIdx, "y");
					else
						rtn = gecJbeamFormatErr;
				}
				else if (cistrncmp((*link)->buf, "\"posZ\"", 6) != NULL)
				{
					if (strchr(s->cIdx, 'z') == NULL)
						strcat(s->cIdx, "z");
					else
						rtn = gecJbeamFormatErr;
				}
			}
		}
		link = &(*link)->pNext;
	}

	//Got all parameters?
	if (strchr(s->cIdx, 'i') == NULL || (s->cIdx, 'x') == NULL || (s->cIdx, 'y') == NULL || (s->cIdx, 'z') == NULL)
	{
		printf("JBEAM format error\r\n");
		rtn = gecJbeamFormatErr;
	}
	else
	{
		//Dump buffer
		int	fStat = fputs(s->buf, s->fDst);
		if (fStat == EOF)
		{
			printf("Error writing to destination file\r\n");
			if (rtn == 0) rtn = gecFileWrite;
		}
	}

	//Reset stuff
	s->buf[0] = '\0';
	ClearLinkedList(s);

	//Bail
	return(rtn);
}

////////////////////////
// Write link to file //
////////////////////////
int						WriteLink(
FILE*					fDst,
const char* const		pStart,
const char* const		pStop)
{
	int			rtn = gecNoError;
	int			fStat = 0;
	const char*	pHere = pStart;
	while (pHere < pStop && rtn == gecNoError)
	{
		fStat = fputc(*pHere, fDst);
		if (fStat == EOF)
		{
			printf("Error writing to destination file\r\n");
			rtn = gecFileWrite;
		}
		pHere++;
	}
	return(rtn);
}

//////////////////////////////////////
// Process subsequent parameter set //
//////////////////////////////////////
int						ProcParmData(
const ParserStatePtr	s)
{
	int		DEBUG_PAUSE = 0;

	//Return var
	int		rtn = gecNoError;
	int		fStat = 0;

	//Retrieve X Y Z numeric fields
	char	fldData[50];
	int		gotX = 0;
	int		gotY = 0;
	int		gotZ = 0;
	double	dX = 0;
	double	dY = 0;
	double	dZ = 0;
	int		fldIdx = 0;
	int		fldLen = 0;
	char*	pHere = NULL;
	char*	pStart = NULL;
	char*	pStop = NULL;
	B2LinkHndl	link = &s->pLinks;
	while (link != s->hNextLink && rtn == gecNoError)
	{
		if (GetLinkBufBounds(s, (*link), &pStart, &pStop) != 0)
		{
			//DEBUG: Check for "98"
			if (cistrncmp(pStart, "\"98\"", 4) == 1 && DEBUG_PAUSE == 0)
			{
				DEBUG_PAUSE = 1;
			}

			//Process field data
			if ((*link)->type == b2Type_FieldData)
			{
				fldLen = int(pStop - pStart);
				strncpy(fldData, pStart, size_t(fldLen));
				fldData[fldLen] = '\0';
				switch (s->cIdx[fldIdx])
				{
					case 'x':	dX = atof(fldData);	gotX = 1;	break;
					case 'y':	dY = atof(fldData);	gotY = 1;	break;
					case 'z':	dZ = atof(fldData);	gotZ = 1;	break;
				}

				//Next data field
				fldIdx++;
			}
		}

		//Next link
		link = &(*link)->pNext;
	}

	//Don't have all required fields
	if (!gotX || !gotY || !gotZ)
	{
		printf("JBEAM format error\r\n");
		rtn = gecJbeamFormatErr;
	}
	else
	{
		///////////////////////////
		// Transform coordinates //
		///////////////////////////
		FloatVecType	v1 = {dX, dY, dZ};
		FloatVecType	v2 = {0, 0, 0};
		FloatVecType	vDist = {s->xRot, s->yRot, s->zRot};
		CoordRot(&v1, &v2, &vDist, s->sRotord);
		dX = v2.x + s->xMov;
		dY = v2.y + s->yMov;
		dZ = v2.z + s->zMov;

		//Loop through links
		pHere = NULL;
		pStart = NULL;
		pStop = NULL;
		fldIdx = 0;
		link = &s->pLinks;
		while (link != s->hNextLink && rtn == gecNoError)
		{
			if (GetLinkBufBounds(s, (*link), &pStart, &pStop) != 0)
			{
				//Process field data
				if ((*link)->type == b2Type_FieldData)
				{
					if(s->cIdx[fldIdx] == 'x')
					{
						sprintf(fldData, "%#.3f", dX);
						rtn = StrFile(s->fDst, fldData);
					}
					else if (s->cIdx[fldIdx] == 'y')
					{
						sprintf(fldData, "%#.3f", dY);
						rtn = StrFile(s->fDst, fldData);
					}
					else if (s->cIdx[fldIdx] == 'z')
					{
						sprintf(fldData, "%#.3f", dZ);
						rtn = StrFile(s->fDst, fldData);
					}
					else
						rtn = WriteLink(s->fDst, pStart, pStop);

					fldIdx++;
				}

				//Process comment or delimiter
				else
				{
					rtn = WriteLink(s->fDst, pStart, pStop);
				}
			}

			//Next link
			link = &(*link)->pNext;
		}
	}

	//Reset stuff
	s->buf[0] = '\0';
	ClearLinkedList(s);

	//Bail
	return(rtn);
}

////////////////////////////
// Node parameter handler //
////////////////////////////
int						ParameterHandler(
const ParserStatePtr	s,
char					c)
{
	//Return var
	int		rtn = gecNoError;

	//Deal with non-comment forward slash
	if (s->lastChar == '/' && c != '/' && c != '*')
	{
		rtn = SaveChar(s, '/', b2Type_FieldData);
	}

	//Deal with received char
	switch (c)
	{
		//Quoted string stuff
		case '\"':
			s->iState |= flagQuote;
			rtn = SaveChar(s, '\"', b2Type_FieldData);
		break;
		
		//Square brackets
		case '[':
			strcatc(s->sState, '[');
			rtn = SaveChar(s, '[', b2Type_FieldData);
		break;
		case ']':
			if (GetLastBoundChar(s) != '[')
			{
				printf("JBEAM format error\r\n");
				rtn = gecJbeamFormatErr;
			}
			else if (charCount(s->sState, '[') == 2 && charCount(s->sState, '{') == 2)
			{
				rtn = SaveChar(s, ']', b2Type_Delimiter);
				if (rtn == gecNoError)
				{
					//Process subsequent (not first) parameter set
					if (s->iState & flagNot1st)
					{
						rtn = ProcParmData(s);
					}

					//Process first parameter set
					else
					{
						rtn = ProcParmHdr(s);
						s->iState |= flagNot1st;
					}
				}

				//Down 1 bracket level
				DelLastBoundChar(s);
			}
			else
			{
				rtn = SaveChar(s, ']', b2Type_FieldData);

				//Down 1 bracket level
				DelLastBoundChar(s);
			}
		break;

		//Curley braces
		case '{':
			strcatc(s->sState, '{');
			rtn = SaveChar(s, '{', b2Type_FieldData);
		break;
		case '}':
			if (GetLastBoundChar(s) != '{')
			{
				printf("JBEAM format error\r\n");
				rtn = gecJbeamFormatErr;
			}
			else
			{
				rtn = SaveChar(s, '}', b2Type_FieldData);

				//Down 1 brace level
				DelLastBoundChar(s);
			}
		break;

		//Comment stuff
		case '/':
			if (s->lastChar == '/')
			{
				rtn = SaveChar(s, '/', b2Type_Comment);
				if(rtn == 0) rtn = SaveChar(s, '/', b2Type_Comment);
				s->iState |= flagCmnt1;
			}
			break;
		case '*':
			if (s->lastChar == '/')
			{
				rtn = SaveChar(s, '/', b2Type_Comment);
				if(rtn == 0) rtn = SaveChar(s, '*', b2Type_Comment);
				s->iState |= flagCmnt2;
			}
			else
				rtn = SaveChar(s, '*', b2Type_FieldData);
		break;
		default:

			//Non-printable chars, space, and comma are delimiters
			if (c <= 32 || c >= 127 || c == ' ' || c == ',')
				rtn = SaveChar(s, c, b2Type_Delimiter);

			//The rest is field data
			else
				rtn = SaveChar(s, c, b2Type_FieldData);
	}

	//Bail
	return(rtn);
}

/////////////////////
// General handler //
/////////////////////
int						GeneralHandler(
const ParserStatePtr	s,
char					c)
{
	//Return var
	int		rtn = SaveChar(s, c, b2Type_FieldData);

	//Act on received char
	switch (c)
	{
		//Double quotation mark
		case '\"':
			s->iState |= flagQuote;
		break;

		//Square brackets
		case '[':
			strcatc(s->sState, '[');
		break;
		case ']':
			if (GetLastBoundChar(s) != '[')
			{
				printf("JBEAM format error\r\n");
				rtn = gecJbeamFormatErr;
			}
			else
			{
				s->iState &= ~(flagNodes | flagNot1st);
				s->cIdx[0] = '\0';
				s->cIdx[1] = '\0';
				s->cIdx[2] = '\0';
				s->cIdx[3] = '\0';
				s->cIdx[4] = '\0';
				DelLastBoundChar(s);
			}
		break;

		//Curley braces
		case '{':
			strcatc(s->sState, '{');
		break;
		case '}':
			if (GetLastBoundChar(s) != '{')
			{
				printf("JBEAM format error\r\n");
				rtn = gecJbeamFormatErr;
			}
			else
				DelLastBoundChar(s);
		break;

		//Comment stuff
		case '/':
			if (s->lastChar == '/')
				s->iState |= flagCmnt1;
		break;
		case '*':
			if (s->lastChar == '/')
				s->iState |= flagCmnt2;
		break;
	}

	//Bail
	return(rtn);
}

/////////////////////////////
// Look for "nodes" string //
/////////////////////////////
void					SeekNodes(
const ParserStatePtr	s)
{
	//Got "nodes"
	if (cistrcmp(s->buf, "nodes") != 0)
	{
		//Not within brackets or too many braces
		if (charCount(s->sState, '[') == 0 &&
			charCount(s->sState, '{') == 2)
		{
			//Standby for parameter collection
			s->iState |= flagNodes;
			s->iState &= ~flagNot1st;
		}
	}
}

////////////////////
// String handler //
////////////////////
int						StringHandler(
const ParserStatePtr	s,
char					c)
{
	//Save character
	int	rtn = SaveChar(s, c, b2Type_FieldData);
	char	str[] = {'\0', '\0'};

	//Act on char
	switch (c)
	{
		//Terminate string
		case '\"':
			s->iState &= ~flagQuote;

			//Check if this is "nodes" string we're looking for
			if (!(s->iState & flagNodes))
			{
				SeekNodes(s);
				s->buf[0] = '\0';
			}
		break;
		default:
			if (!(s->iState & flagNodes))
			{
				str[0] = c;
				strcat(s->buf, str);
			}
	}

	//Bail
	return(rtn);
}

/////////////////////
// Comment handler //
/////////////////////
int						CommentHandler(
const ParserStatePtr	s,
char					c)
{
	//Link type
	B2LinkTypeEnum	lt = b2Type_Comment;
	
	//Act on char
	switch (c)
	{
		//Terminate single-line comment
		case '\r':
		case '\n':
			if (s->iState & flagCmnt1)
				s->iState &= ~flagCmnt1;
			lt = b2Type_Delimiter;
		break;

		//Terminate multi-line comment
		case '/':
			if (s->lastChar == '*' && (s->iState & flagCmnt2))
				s->iState &= ~flagCmnt2;
		break;
	}

	//Save character
	int	rtn = SaveChar(s, c, lt);

	//Bail
	return(rtn);
}

////////////////////////////////////////
// Dispatch to handler based on state //
////////////////////////////////////////
int						DispatchChar(
const ParserStatePtr	s,
char					c)
{
	//Vars
	int	rtn = gecNoError;

	//Collecting comment chars
	if (s->iState & (flagCmnt1 | flagCmnt2))
		rtn = CommentHandler(s, c);

	//Colecting string chars
	else if (s->iState & flagQuote)
		rtn = StringHandler(s, c);

	//Collecting parameters
	else if (InParmCol(s))
		rtn = ParameterHandler(s, c);

	//Handle characters regularly
	else
		rtn = GeneralHandler(s, c);

	//Save last char processed
	s->lastChar = c;

	//Bail
	return(rtn);
}

//////////////////////
// Parse JBEAM File //
//////////////////////
int						ParseJbeam(
const ParserStatePtr	s,
FILE*					fSrc)
{
	//Return var
	int rtn = gecNoError;

	//Loop through all characters in source file
	int c = 0;
	do
	{
		c = fgetc(fSrc);
		if (c != EOF)
		{
			//Process character
			rtn = DispatchChar(s, char(c));
		}
	}
	while (c != EOF && rtn == gecNoError);

	//Error reading file?
	if (ferror(fSrc) != 0)
	{
		printf("Error reading from source file\r\n");
		rtn = gecFileRead;
	}

	//Bail
	return(rtn);
}

//////////////////////////
// Transform JBEAM File //
//////////////////////////
int				xFormJbeam(
char* const		sJbeam,
char* const		sRotord,
double			xRot,
double			yRot,
double			zRot,
double			xMov,
double			yMov,
double			zMov)
{
	//Return var
	int	rtn = gecNoError;

	//Parser state
	ParserStatePtr	Pstate = ParserStatePtr(malloc(sizeof(ParserStateType)));
	if (Pstate == NULL)
	{
		printf("Cannot allocate state structure\r\n");
		rtn = gecMemAllocErr;
	}
	else
	{
		memset(Pstate, 0, sizeof(ParserStateType));
		Pstate->hNextLink = &Pstate->pLinks;

		//Set input parameters
		Pstate->sRotord = sRotord;
		Pstate->xRot = xRot;
		Pstate->yRot = yRot;
		Pstate->zRot = zRot;
		Pstate->xMov = xMov;
		Pstate->yMov = yMov;
		Pstate->zMov = zMov;

		//Allocate primary buffer
		Pstate->buf = (char*)malloc(4096);
		if (Pstate->buf == NULL)
		{
			printf("Cannot allocate primary buffer\r\n");
			rtn = gecMemAllocErr;
		}
		else
		{
			Pstate->buf[0] = '\0';

			//Allocate state buffer;
			Pstate->sState = (char*)malloc(128);
			if (Pstate->sState == NULL)
			{
				printf("Cannot allocate state buffer\r\n");
				rtn = gecMemAllocErr;
			}
			else
			{
				Pstate->sState[0] = '\0';

				//Construct destination file name
				char	sDst[FILENAME_MAX];
				rtn = GetDestFileName(sJbeam, sDst);

				//Open JBEAM source file
				FILE* fSrc = fopen(sJbeam, "r");
				if (fSrc == NULL)
				{
					printf("Cannot open read file \"%s\"\r\n", sJbeam);
					rtn = gecFileNonexistent;
				}
				else
				{
					//Open transformed JBEAM destination file
					Pstate->fDst = fopen(sDst, "w");
					if (Pstate->fDst == NULL)
					{
						printf("Cannot open write file \"%s\"\r\n", sDst);
						return(gecFileCantMake);
					}
					else
					{
						//Parse JBEAM
						rtn = ParseJbeam(Pstate, fSrc);

						//Check for unbuffered forward slash
						if (Pstate->lastChar == '/' && charCount(Pstate->sState, '[') >= 2)
						{
							strcat(Pstate->buf, "/");
						}

						//Got links; Dump them
						if (Pstate->pLinks != NULL)
						{
							int fStat = fputs(Pstate->buf, Pstate->fDst);
							if (fStat == EOF)
							{
								printf("Error writing to destination file\r\n");
								if (rtn == 0) rtn = gecFileWrite;
							}
							ClearLinkedList(Pstate);
						}

						//Close file
						fclose(Pstate->fDst);
						Pstate->fDst = NULL;
					}

					//Close file
					fclose(fSrc);
					fSrc = NULL;
				}

				//Free memory
				free(Pstate->sState);
				Pstate->sState = NULL;
			}

			//Free memory
			free(Pstate->buf);
			Pstate->buf = NULL;
		}

		//Free memory
		free(Pstate);
		Pstate = NULL;
	}

	//Bail
	return(rtn);
}
